# Copyright (C) 2019-2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import Counter
import datetime
from functools import partial
import logging
import traceback
from typing import Any, Callable
from typing import Counter as CounterT
from typing import Dict, List, Optional, Tuple, TypeVar, Union, cast
from uuid import uuid4

import yaml

try:
    from systemd.daemon import notify
except ImportError:
    notify = None

from swh.core.statsd import statsd
from swh.journal.serializers import kafka_to_value
from swh.model.hashutil import hash_to_hex
from swh.model.model import (
    BaseContent,
    BaseModel,
    Content,
    Directory,
    ExtID,
    HashableObject,
    MetadataAuthority,
    MetadataFetcher,
    ModelObjectType,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    RawExtrinsicMetadata,
    Release,
    Revision,
    SkippedContent,
    Snapshot,
)
from swh.storage.exc import HashCollision, StorageArgumentException
from swh.storage.interface import StorageInterface
from swh.storage.utils import remove_keys

logger = logging.getLogger(__name__)

GRAPH_OPERATIONS_METRIC = "swh_graph_replayer_operations_total"
GRAPH_DURATION_METRIC = "swh_graph_replayer_duration_seconds"


OBJECT_CONVERTERS: Dict[ModelObjectType, Callable[[Dict], BaseModel]] = {
    Origin.object_type: Origin.from_dict,
    OriginVisit.object_type: OriginVisit.from_dict,
    OriginVisitStatus.object_type: OriginVisitStatus.from_dict,
    Snapshot.object_type: Snapshot.from_dict,
    Revision.object_type: Revision.from_dict,
    Release.object_type: Release.from_dict,
    Directory.object_type: Directory.from_dict,
    Content.object_type: Content.from_dict,
    SkippedContent.object_type: SkippedContent.from_dict,
    MetadataAuthority.object_type: MetadataAuthority.from_dict,
    MetadataFetcher.object_type: MetadataFetcher.from_dict,
    RawExtrinsicMetadata.object_type: RawExtrinsicMetadata.from_dict,
    ExtID.object_type: ExtID.from_dict,
}
# Deprecated, for BW compat only.
object_converter_fn = OBJECT_CONVERTERS


OBJECT_FIXERS = {
    # drop the metadata field from the revision or visit status (if any); this field is
    # about to be dropped from the data model (in favor of
    # raw_extrinsic_metadata) and there can be bogus values in the existing
    # journal (metadata with \0000 in it)
    Revision.object_type: partial(remove_keys, keys=("metadata",)),
    OriginVisitStatus.object_type: partial(remove_keys, keys=("metadata",)),
}


def now():
    return datetime.datetime.now(datetime.UTC)


class ModelObjectDeserializer:
    """A swh.journal object deserializer that checks object validity and reports
    invalid objects

    The deserializer will directly produce BaseModel objects from journal
    objects representations.

    If validation is activated and the object is hashable, it will check if the
    computed hash matches the identifier of the object.

    If the object is invalid and a 'reporter' function is given, it will be
    called with 2 arguments::

      reporter(object_id, journal_msg)

    Where 'object_id' is a string representation of the object identifier (from
    the journal message), and 'journal_msg' is the row message (bytes)
    retrieved from the journal.

    If 'raise_on_error' is True, a 'StorageArgumentException' exception is
    raised.

    If 'known_mismatched_hashes' is given, it must be a tuple of triplets
    (object_type, object_id, expected_id) listing objects that store invalid
    hash (object_id) instead of the computed expected_id, but should not be
    discarded (i.e. they should be replicated by the replayer despite being
    invalid).

    Typical usage::

      deserializer = ModelObjectDeserializer(validate=True, reporter=reporter_cb)
      client = get_journal_client(
          cls="kafka", value_deserializer=deserializer, **cfg)

    """

    def __init__(
        self,
        validate: bool = True,
        raise_on_error: bool = False,
        reporter: Optional[Callable[[str, bytes], None]] = None,
        known_mismatched_hashes: Optional[Tuple[Tuple[str, bytes, bytes]]] = None,
    ):
        self.validate = validate
        self.reporter = reporter
        self.raise_on_error = raise_on_error
        self.known_mismatched_hashes = known_mismatched_hashes

    def convert(self, object_type: str, msg: bytes) -> Optional[BaseModel]:
        dict_repr = kafka_to_value(msg)
        obj_type = ModelObjectType(object_type)
        if obj_type in OBJECT_FIXERS:
            dict_repr = OBJECT_FIXERS[obj_type](dict_repr)
        try:
            obj = OBJECT_CONVERTERS[obj_type](dict_repr)
        except ValueError as exc:
            # we do not catch AttributeTypeError here since these are (most
            # likely) a clue of something very wrong is occurring, so better crash
            error_msg = f"Unable to create model object {object_type}: {repr(exc)}"
            self.report_failure(msg, (object_type, dict_repr), exc)
            if self.raise_on_error:
                raise StorageArgumentException(error_msg)
            return None

        if self.validate:
            if isinstance(obj, HashableObject):
                cid = obj.compute_hash()
                if obj.id != cid:
                    error_msg = (
                        f"Object has id {hash_to_hex(obj.id)}, "
                        f"but it should be {hash_to_hex(cid)}: {obj}."
                    )
                    if (
                        self.known_mismatched_hashes is None
                        or (obj_type.value, obj.id, cid)
                        not in self.known_mismatched_hashes
                    ):
                        logger.error(error_msg)
                        self.report_failure(msg, obj, None)
                        if self.raise_on_error:
                            raise StorageArgumentException(error_msg)
                        return None
                    else:
                        logger.warning(
                            error_msg + " Known exception, replicating it anyway."
                        )
        return obj

    def report_failure(
        self,
        msg: bytes,
        obj: Union[BaseModel, Tuple[str, Dict[str, Any]]],
        exc: BaseException | None,
    ):
        if self.reporter:
            oid: str
            if isinstance(obj, tuple):
                object_type, dict_repr = obj
                if "id" in dict_repr:
                    uid = dict_repr["id"]
                    assert isinstance(uid, bytes)
                    oid = f"{object_type}:{uid.hex()}"
                else:
                    oid = f"{object_type}:uuid:{uuid4()}"
            else:
                dict_repr = obj.to_dict()
                if isinstance(obj, HashableObject):
                    oid = f"{obj.object_type}:{obj.id.hex()}"
                else:
                    oid = f"{obj.object_type}:uuid:{uuid4()}"

            key = f"{now().isoformat()}/{oid}"
            value = {
                "kafka_message": msg,
                "obj": dict_repr,
                "exc": traceback.format_exception(exc) if exc is not None else None,
            }
            self.reporter(key, yaml.dump(value).encode())


def process_replay_objects(
    all_objects: Dict[str, List[BaseModel]], *, storage: StorageInterface
) -> None:
    for object_type, objects in all_objects.items():
        logger.debug("Inserting %s %s objects", len(objects), object_type)
        with statsd.timed(GRAPH_DURATION_METRIC, tags={"object_type": object_type}):
            _insert_objects(object_type, objects, storage)
            storage.flush()
        statsd.increment(
            GRAPH_OPERATIONS_METRIC, len(objects), tags={"object_type": object_type}
        )
    if notify:
        notify("WATCHDOG=1")


ContentType = TypeVar("ContentType", bound=BaseContent)


def collision_aware_content_add(
    contents: List[ContentType],
    content_add_fn: Callable[[List[ContentType]], Dict[str, int]],
) -> Dict[str, int]:
    """Add contents to storage. If a hash collision is detected, an error is
       logged. Then this adds the other non colliding contents to the storage.

    Args:
        content_add_fn: Storage content callable
        contents: List of contents or skipped contents to add to storage

    """
    if not contents:
        return {}
    colliding_content_hashes: List[Dict[str, Any]] = []
    results: CounterT[str] = Counter()
    while True:
        try:
            results.update(content_add_fn(contents))
        except HashCollision as e:
            logger.debug("Hash collision detected: %s", e)
            colliding_content_hashes.append(
                {
                    "algo": e.algo,
                    "hash": e.hash_id,  # hex hash id
                    "objects": e.colliding_contents,  # hex hashes
                }
            )
            colliding_hashes = e.colliding_content_hashes()
            logger.debug("Dropping colliding hashes: %s", colliding_hashes)

            # Drop the colliding contents from the transaction
            contents = [c for c in contents if c.hashes() not in colliding_hashes]
        else:
            # Successfully added contents, we are done
            break
    if colliding_content_hashes:
        for collision in colliding_content_hashes:
            logger.error("Collision detected: %(collision)s", {"collision": collision})
    return dict(results)


def _insert_objects(
    object_type: str, objects: List[BaseModel], storage: StorageInterface
) -> None:
    """Insert objects of type object_type in the storage."""
    obj_type = ModelObjectType(object_type)
    if obj_type not in OBJECT_CONVERTERS:
        logger.warning("Received a series of %s, this should not happen", object_type)
        return

    method = getattr(storage, f"{object_type}_add")
    if obj_type == SkippedContent.object_type:
        method = partial(collision_aware_content_add, content_add_fn=method)
    elif obj_type == Content.object_type:
        method = partial(
            collision_aware_content_add, content_add_fn=storage.content_add_metadata
        )
    elif obj_type in (
        OriginVisit.object_type,
        OriginVisitStatus.object_type,
    ):
        origins: List[Origin] = []
        for obj in cast(List[Union[OriginVisit, OriginVisitStatus]], objects):
            origins.append(Origin(url=obj.origin))
        storage.origin_add(origins)
    elif obj_type == RawExtrinsicMetadata.object_type:
        emds = cast(List[RawExtrinsicMetadata], objects)
        authorities = {emd.authority for emd in emds}
        fetchers = {emd.fetcher for emd in emds}
        storage.metadata_authority_add(list(authorities))
        storage.metadata_fetcher_add(list(fetchers))
    method(objects)



# Copyright (C) 2017-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Storage backfiller.

The backfiller goal is to produce back part or all of the objects
from a storage to the journal topics

Current implementation consists in the JournalBackfiller class.

It simply reads the objects from the storage and sends every object identifier back to
the journal.

"""

import logging
from typing import Any, Callable, Dict, Iterable, Iterator, Optional, Tuple, Union

from swh.core.db import BaseDb
from swh.model.model import (
    BaseModel,
    Content,
    Directory,
    DirectoryEntry,
    ExtID,
    MetadataAuthority,
    MetadataFetcher,
    ModelObjectType,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    RawExtrinsicMetadata,
    Release,
    Revision,
    SkippedContent,
    Snapshot,
    SnapshotBranch,
    SnapshotTargetType,
)
from swh.model.swhids import ExtendedObjectType
from swh.storage.postgresql.converters import (
    db_to_extid,
    db_to_raw_extrinsic_metadata,
    db_to_release,
    db_to_revision,
)
from swh.storage.replay import OBJECT_CONVERTERS
from swh.storage.writer import JournalWriter

logger = logging.getLogger(__name__)

PARTITION_KEY = {
    Content.object_type: "sha1",
    SkippedContent.object_type: "sha1",
    Directory.object_type: "id",
    ExtID.object_type: "target",
    MetadataAuthority.object_type: "type, url",
    MetadataFetcher.object_type: "name, version",
    RawExtrinsicMetadata.object_type: "target",
    Revision.object_type: "revision.id",
    Release.object_type: "release.id",
    Snapshot.object_type: "id",
    Origin.object_type: "id",
    OriginVisit.object_type: "origin_visit.origin",
    OriginVisitStatus.object_type: "origin_visit_status.origin",
}

COLUMNS = {
    Content.object_type: [
        "sha1",
        "sha1_git",
        "sha256",
        "blake2s256",
        "length",
        "status",
        "ctime",
    ],
    SkippedContent.object_type: [
        "sha1",
        "sha1_git",
        "sha256",
        "blake2s256",
        "length",
        "ctime",
        "status",
        "reason",
    ],
    Directory.object_type: [
        "id",
        "dir_entries",
        "file_entries",
        "rev_entries",
        "raw_manifest",
    ],
    ExtID.object_type: [
        "extid_type",
        "extid",
        "extid_version",
        "target_type",
        "target",
    ],
    MetadataAuthority.object_type: ["type", "url"],
    MetadataFetcher.object_type: ["name", "version"],
    Origin.object_type: ["url"],
    OriginVisit.object_type: [
        "visit",
        "type",
        ("origin.url", "origin"),
        "date",
    ],
    OriginVisitStatus.object_type: [
        ("origin_visit_status.visit", "visit"),
        ("origin.url", "origin"),
        ("origin_visit_status.date", "date"),
        "type",
        "snapshot",
        "status",
    ],
    RawExtrinsicMetadata.object_type: [
        "raw_extrinsic_metadata.type",
        "raw_extrinsic_metadata.target",
        "metadata_authority.type",
        "metadata_authority.url",
        "metadata_fetcher.name",
        "metadata_fetcher.version",
        "discovery_date",
        "format",
        "raw_extrinsic_metadata.metadata",
        "origin",
        "visit",
        "snapshot",
        "release",
        "revision",
        "path",
        "directory",
    ],
    Revision.object_type: [
        ("revision.id", "id"),
        "date",
        "date_offset_bytes",
        "committer_date",
        "committer_date_offset_bytes",
        "type",
        "directory",
        "message",
        "synthetic",
        "extra_headers",
        (
            "array(select parent_id::bytea from revision_history rh "
            "where rh.id = revision.id order by rh.parent_rank asc)",
            "parents",
        ),
        "raw_manifest",
        ("a.id", "author_id"),
        ("a.name", "author_name"),
        ("a.email", "author_email"),
        ("a.fullname", "author_fullname"),
        ("c.id", "committer_id"),
        ("c.name", "committer_name"),
        ("c.email", "committer_email"),
        ("c.fullname", "committer_fullname"),
    ],
    Release.object_type: [
        ("release.id", "id"),
        "date",
        "date_offset_bytes",
        "comment",
        ("release.name", "name"),
        "synthetic",
        "target",
        "target_type",
        ("a.id", "author_id"),
        ("a.name", "author_name"),
        ("a.email", "author_email"),
        ("a.fullname", "author_fullname"),
        "raw_manifest",
    ],
    Snapshot.object_type: ["id", "object_id"],
}


JOINS = {
    Release.object_type: ["person a on release.author=a.id"],
    Revision.object_type: [
        "person a on revision.author=a.id",
        "person c on revision.committer=c.id",
    ],
    OriginVisit.object_type: ["origin on origin_visit.origin=origin.id"],
    OriginVisitStatus.object_type: [
        "origin on origin_visit_status.origin=origin.id",
    ],
    RawExtrinsicMetadata.object_type: [
        "metadata_authority on "
        "raw_extrinsic_metadata.authority_id=metadata_authority.id",
        "metadata_fetcher on raw_extrinsic_metadata.fetcher_id=metadata_fetcher.id",
    ],
}

EXTRA_WHERE = {
    # hack to force the right index usage on table extid
    ExtID.object_type: "target_type in ('revision', 'release', 'content', 'directory')"
}


def directory_converter(db: BaseDb, directory_d: Dict[str, Any]) -> Directory:
    """Convert directory from the flat representation to swh model
    compatible objects.

    """
    columns = ["target", "name", "perms"]
    query_template = """
    select %(columns)s
    from directory_entry_%(type)s
    where id = Any(%%s)
    """

    types = ["file", "dir", "rev"]

    entries = []
    with db.cursor() as cur:
        for type in types:
            ids = directory_d.pop("%s_entries" % type)
            if not ids:
                continue
            query = query_template % {
                "columns": ",".join(columns),
                "type": type,
            }
            cur.execute(query, (list(ids),))
            for row in cur:
                entry_d = dict(zip(columns, row))
                entry = DirectoryEntry(
                    name=entry_d["name"],
                    type=type,
                    target=entry_d["target"],
                    perms=entry_d["perms"],
                )
                entries.append(entry)

    (is_corrupt, dir_) = Directory.from_possibly_duplicated_entries(
        id=directory_d["id"],
        entries=tuple(entries),
        raw_manifest=directory_d["raw_manifest"],
    )
    if is_corrupt:
        logger.info("%s has duplicated entries", dir_.swhid())

    return dir_


def raw_extrinsic_metadata_converter(
    db: BaseDb, metadata: Dict[str, Any]
) -> RawExtrinsicMetadata:
    """Convert a raw extrinsic metadata from the flat representation to swh model
    compatible objects.

    """
    return db_to_raw_extrinsic_metadata(metadata)


def extid_converter(db: BaseDb, extid: Dict[str, Any]) -> ExtID:
    """Convert an extid from the flat representation to swh model
    compatible objects.

    """
    return db_to_extid(extid)


def revision_converter(db: BaseDb, revision_d: Dict[str, Any]) -> Revision:
    """Convert revision from the flat representation to swh model
    compatible objects.

    """
    return db_to_revision(revision_d)


def release_converter(db: BaseDb, release_d: Dict[str, Any]) -> Release:
    """Convert release from the flat representation to swh model
    compatible objects.

    """
    return db_to_release(release_d)


def snapshot_converter(db: BaseDb, snapshot_d: Dict[str, Any]) -> Snapshot:
    """Convert snapshot from the flat representation to swh model
    compatible objects.

    """
    columns = ["name", "target", "target_type"]
    query = """
    select %s
    from snapshot_branches sbs
    inner join snapshot_branch sb on sb.object_id=sbs.branch_id
    where sbs.snapshot_id=%%s
    """ % ", ".join(
        columns
    )
    with db.cursor() as cur:
        cur.execute(query, (snapshot_d["object_id"],))
        branches = {}
        for name, *row in cur:
            branch_d = dict(zip(columns[1:], row))
            if branch_d["target"] is not None and branch_d["target_type"] is not None:
                branch: Optional[SnapshotBranch] = SnapshotBranch(
                    target=branch_d["target"],
                    target_type=SnapshotTargetType(branch_d["target_type"]),
                )
            else:
                branch = None
            branches[name] = branch

    return Snapshot(
        id=snapshot_d["id"],
        branches=branches,
    )


CONVERTERS: Dict[ModelObjectType, Callable[[BaseDb, Dict[str, Any]], BaseModel]] = {
    Directory.object_type: directory_converter,
    ExtID.object_type: extid_converter,
    RawExtrinsicMetadata.object_type: raw_extrinsic_metadata_converter,
    Revision.object_type: revision_converter,
    Release.object_type: release_converter,
    Snapshot.object_type: snapshot_converter,
}


def object_to_offset(object_id, numbits):
    """Compute the index of the range containing object id, when dividing
       space into 2^numbits.

    Args:
        object_id (str): The hex representation of object_id
        numbits (int): Number of bits in which we divide input space

    Returns:
        The index of the range containing object id

    """
    q, r = divmod(numbits, 8)
    length = q + (r != 0)
    shift_bits = 8 - r if r else 0

    truncated_id = object_id[: length * 2]
    if len(truncated_id) < length * 2:
        truncated_id += "0" * (length * 2 - len(truncated_id))

    truncated_id_bytes = bytes.fromhex(truncated_id)
    return int.from_bytes(truncated_id_bytes, byteorder="big") >> shift_bits


def byte_ranges(
    numbits: int, start_object: Optional[str] = None, end_object: Optional[str] = None
) -> Iterator[Tuple[Optional[bytes], Optional[bytes]]]:
    """Generate start/end pairs of bytes spanning numbits bits and
       constrained by optional start_object and end_object.

    Args:
        numbits: Number of bits in which we divide input space
        start_object: Hex object id contained in the first range
                            returned
        end_object: Hex object id contained in the last range
                          returned

    Yields:
        2^numbits pairs of bytes

    """
    q, r = divmod(numbits, 8)
    length = q + (r != 0)
    shift_bits = 8 - r if r else 0

    def to_bytes(i):
        return int.to_bytes(i << shift_bits, length=length, byteorder="big")

    start_offset = 0
    end_offset = 1 << numbits

    if start_object is not None:
        start_offset = object_to_offset(start_object, numbits)
    if end_object is not None:
        end_offset = object_to_offset(end_object, numbits) + 1

    for start in range(start_offset, end_offset):
        end = start + 1

        if start == 0:
            yield None, to_bytes(end)
        elif end == 1 << numbits:
            yield to_bytes(start), None
        else:
            yield to_bytes(start), to_bytes(end)


def raw_extrinsic_metadata_target_ranges(
    start_object: Optional[str] = None, end_object: Optional[str] = None
) -> Iterator[Tuple[Optional[str], Optional[str]]]:
    """Generate ranges of values for the `target` attribute of `raw_extrinsic_metadata`
    objects.

    This generates one range for all values before the first SWHID (which would
    correspond to raw origin URLs), then a number of hex-based ranges for each
    known type of SWHID (2**12 ranges for directories, 2**8 ranges for all other
    types). Finally, it generates one extra range for values above all possible
    SWHIDs.
    """
    if start_object is None:
        start_object = ""

    swhid_target_types = sorted(type.value for type in ExtendedObjectType)

    first_swhid = f"swh:1:{swhid_target_types[0]}:"

    # Generate a range for url targets, if the starting object is before SWHIDs
    if start_object < first_swhid:
        yield start_object, (
            first_swhid
            if end_object is None or end_object >= first_swhid
            else end_object
        )

    if end_object is not None and end_object <= first_swhid:
        return

    # Prime the following loop, which uses the upper bound of the previous range
    # as lower bound, to account for potential targets between two valid types
    # of SWHIDs (even though they would eventually be rejected by the
    # RawExtrinsicMetadata parser, they /might/ exist...)
    end_swhid = first_swhid

    # Generate ranges for swhid targets
    for target_type in swhid_target_types:
        finished = False
        base_swhid = f"swh:1:{target_type}:"
        last_swhid = base_swhid + ("f" * 40)

        if start_object > last_swhid:
            continue

        # Generate 2**8 or 2**12 ranges
        for _, end in byte_ranges(12 if target_type == "dir" else 8):
            # Reuse previous upper bound
            start_swhid = end_swhid

            # Use last_swhid for this object type if on the last byte range
            end_swhid = (base_swhid + end.hex()) if end is not None else last_swhid

            # Ignore out of bounds ranges
            if start_object >= end_swhid:
                continue

            # Potentially clamp start of range to the first object requested
            start_swhid = max(start_swhid, start_object)

            # Handle ending the loop early if the last requested object id is in
            # the current range
            if end_object is not None and end_swhid >= end_object:
                end_swhid = end_object
                finished = True

            yield start_swhid, end_swhid

            if finished:
                return

    # Generate one final range for potential raw origin URLs after the last
    # valid SWHID
    start_swhid = max(start_object, end_swhid)
    yield start_swhid, end_object


def integer_ranges(
    start: str, end: str, block_size: int = 1000
) -> Iterator[Tuple[Optional[int], Optional[int]]]:
    for range_start in range(int(start), int(end), block_size):
        if range_start == 0:
            yield None, block_size
        elif range_start + block_size > int(end):
            yield range_start, int(end)
        else:
            yield range_start, range_start + block_size


RANGE_GENERATORS: Dict[
    ModelObjectType,
    Union[
        Callable[[str, str], Iterable[Tuple[Optional[str], Optional[str]]]],
        Callable[[str, str], Iterable[Tuple[Optional[bytes], Optional[bytes]]]],
        Callable[[str, str], Iterable[Tuple[Optional[int], Optional[int]]]],
    ],
] = {
    Content.object_type: lambda start, end: byte_ranges(24, start, end),
    SkippedContent.object_type: lambda start, end: [(None, None)],
    Directory.object_type: lambda start, end: byte_ranges(24, start, end),
    ExtID.object_type: lambda start, end: byte_ranges(24, start, end),
    Revision.object_type: lambda start, end: byte_ranges(24, start, end),
    Release.object_type: lambda start, end: byte_ranges(16, start, end),
    RawExtrinsicMetadata.object_type: raw_extrinsic_metadata_target_ranges,
    Snapshot.object_type: lambda start, end: byte_ranges(16, start, end),
    Origin.object_type: integer_ranges,
    OriginVisit.object_type: integer_ranges,
    OriginVisitStatus.object_type: integer_ranges,
}


def compute_query(obj_type, start, end):
    object_type = ModelObjectType(obj_type)
    columns = COLUMNS.get(object_type)
    join_specs = JOINS.get(object_type, [])
    join_clause = "\n".join("left join %s" % clause for clause in join_specs)
    additional_where = EXTRA_WHERE.get(object_type)

    where = []
    where_args = []
    if start:
        where.append("%(keys)s >= %%s")
        where_args.append(start)
    if end:
        where.append("%(keys)s < %%s")
        where_args.append(end)

    if additional_where:
        where.append(additional_where)

    where_clause = ""
    if where:
        where_clause = ("where " + " and ".join(where)) % {
            "keys": "(%s)" % PARTITION_KEY[object_type]
        }

    column_specs = []
    column_aliases = []
    for column in columns:
        if isinstance(column, str):
            column_specs.append(column)
            column_aliases.append(column)
        else:
            column_specs.append("%s as %s" % column)
            column_aliases.append(column[1])

    query = """
select %(columns)s
from %(table)s
%(join)s
%(where)s
    """ % {
        "columns": ",".join(column_specs),
        "table": obj_type,
        "join": join_clause,
        "where": where_clause,
    }

    return query, where_args, column_aliases


def fetch(db, obj_type, start, end):
    """Fetch all obj_type's identifiers from db.

    This opens one connection, stream objects and when done, close
    the connection.

    Args:
        db (BaseDb): Db connection object
        obj_type (str): Object type
        start (Union[bytes|Tuple]): Range start identifier
        end (Union[bytes|Tuple]): Range end identifier

    Raises:
        ValueError if obj_type is not supported

    Yields:
        Objects in the given range

    """
    query, where_args, column_aliases = compute_query(obj_type, start, end)
    object_type = ModelObjectType(obj_type)
    converter = CONVERTERS.get(object_type)
    with db.cursor() as cursor:
        logger.debug("Fetching data for table %s", obj_type)
        logger.debug("query: %s %s", query, where_args)
        cursor.execute(query, where_args)
        for row in cursor:
            record = dict(zip(column_aliases, row))
            if converter:
                record = converter(db, record)
            else:
                record = OBJECT_CONVERTERS[object_type](record)

            logger.debug("record: %s", record)
            yield record


def _format_range_bound(bound):
    if isinstance(bound, bytes):
        return bound.hex()
    else:
        return str(bound)


MANDATORY_KEYS = ["storage", "journal_writer"]


class JournalBackfiller:
    """Class in charge of reading the storage's objects and sends those
    back to the journal's topics.

    This is designed to be run periodically.

    """

    def __init__(self, config=None):
        self.config = config
        self.check_config(config)
        self._db = None
        self.writer = JournalWriter({"cls": "kafka", **self.config["journal_writer"]})
        assert self.writer.journal is not None

    @property
    def db(self):
        if self._db is None:
            self._db = BaseDb.connect(self.config["storage"]["db"])
        return self._db

    def check_config(self, config):
        missing_keys = []
        for key in MANDATORY_KEYS:
            if not config.get(key):
                missing_keys.append(key)

        if missing_keys:
            raise ValueError(
                "Configuration error: The following keys must be"
                " provided: %s" % (",".join(missing_keys),)
            )

        if "cls" not in config["storage"] or config["storage"]["cls"] not in (
            "local",
            "postgresql",
        ):
            raise ValueError(
                "swh storage backfiller must be configured to use a local"
                " (PostgreSQL) storage"
            )

    def parse_arguments(self, object_type, start_object, end_object):
        """Parse arguments

        Raises:
            ValueError for unsupported object type
            ValueError if object ids are not parseable

        Returns:
            Parsed start and end object ids

        """
        try:
            model_object_type = ModelObjectType(object_type)
        except ValueError:
            model_object_type = None
        if model_object_type is None or model_object_type not in COLUMNS:
            raise ValueError(
                "Object type %s is not supported. "
                "The only possible values are %s"
                % (object_type, ", ".join(sorted(key.value for key in COLUMNS)))
            )

        if model_object_type in [
            Origin.object_type,
            OriginVisit.object_type,
            OriginVisitStatus.object_type,
        ]:
            start_object = start_object or "0"
            end_object = end_object or str(100_000_000)  # hard-coded limit

        return start_object, end_object

    def run(self, object_type, start_object, end_object, dry_run=False):
        """Reads storage's subscribed object types and send them to the
        journal's reading topic.

        """
        start_object, end_object = self.parse_arguments(
            object_type, start_object, end_object
        )

        for range_start, range_end in RANGE_GENERATORS[ModelObjectType(object_type)](
            start_object, end_object
        ):
            logger.info(
                "Processing %s range %s to %s",
                object_type,
                _format_range_bound(range_start),
                _format_range_bound(range_end),
            )

            objects = fetch(self.db, object_type, start=range_start, end=range_end)

            if not dry_run:
                self.writer.write_additions(object_type, objects)
            else:
                # only consume the objects iterator to check for any potential
                # decoding/encoding errors
                for obj in objects:
                    pass


if __name__ == "__main__":
    print('Please use the "swh-journal backfiller run" command')



# Copyright (C) 2018-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from copy import deepcopy

from swh.model.hashutil import hash_to_bytes, hash_to_hex
from swh.model.model import Revision
from swh.storage.algos.revisions_walker import get_revisions_walker

# For those tests, we will walk the following revisions history
# with different orderings:
#
# *   commit b364f53155044e5308a0f73abb3b5f01995a5b7d
# |\  Merge: 836d498 b94886c
# | | Author:     Adam <adam.janicki@roche.com>
# | | AuthorDate: Fri Oct 4 12:50:49 2013 +0200
# | | Commit:     Adam <adam.janicki@roche.com>
# | | CommitDate: Fri Oct 4 12:50:49 2013 +0200
# | |
# | |     Merge branch 'release/1.0'
# | |
# | * commit b94886c500c46e32dc3d7ebae8a5409accd592e5
# | | Author:     Adam <adam.janicki@roche.com>
# | | AuthorDate: Fri Oct 4 12:50:38 2013 +0200
# | | Commit:     Adam <adam.janicki@roche.com>
# | | CommitDate: Fri Oct 4 12:50:38 2013 +0200
# | |
# | |     updating poms for 1.0 release
# | |
# | * commit 0cb6b4611d65bee0f57821dac7f611e2f8a02433
# | | Author:     Adam <adam.janicki@roche.com>
# | | AuthorDate: Fri Oct 4 12:50:38 2013 +0200
# | | Commit:     Adam <adam.janicki@roche.com>
# | | CommitDate: Fri Oct 4 12:50:38 2013 +0200
# | |
# | |     updating poms for 1.0 release
# | |
# | * commit 2b0240c6d682bad51532eec15b8a7ed6b75c8d31
# | | Author:     Adam Janicki <janickia>
# | | AuthorDate: Fri Oct 4 12:50:22 2013 +0200
# | | Commit:     Adam Janicki <janickia>
# | | CommitDate: Fri Oct 4 12:50:22 2013 +0200
# | |
# | |     For 1.0 release. Allow untracked.
# | |
# | * commit b401c50863475db4440c85c10ac0b6423b61554d
# | | Author:     Adam <adam.janicki@roche.com>
# | | AuthorDate: Fri Oct 4 12:48:12 2013 +0200
# | | Commit:     Adam <adam.janicki@roche.com>
# | | CommitDate: Fri Oct 4 12:48:12 2013 +0200
# | |
# | |     updating poms for 1.0 release
# | |
# | * commit 9c5051397e5c2e0c258bb639c3dd34406584ca10
# |/  Author:     Adam Janicki <janickia>
# |   AuthorDate: Fri Oct 4 12:47:48 2013 +0200
# |   Commit:     Adam Janicki <janickia>
# |   CommitDate: Fri Oct 4 12:47:48 2013 +0200
# |
# |       For 1.0 release.
# |
# * commit 836d498396fb9b5d45c896885f84d8d60a5651dc
# | Author:     Adam Janicki <janickia>
# | AuthorDate: Fri Oct 4 12:08:16 2013 +0200
# | Commit:     Adam Janicki <janickia>
# | CommitDate: Fri Oct 4 12:08:16 2013 +0200
# |
# |     Add ignores
# |
# * commit ee96c2a2d397b79070d2b6fe3051290963748358
# | Author:     Adam <adam.janicki@roche.com>
# | AuthorDate: Fri Oct 4 10:48:16 2013 +0100
# | Commit:     Adam <adam.janicki@roche.com>
# | CommitDate: Fri Oct 4 10:48:16 2013 +0100
# |
# |     Reset author
# |
# * commit 8f89dda8e072383cf50d42532ae8f52ad89f8fdf
#   Author:     Adam <adam.janicki@roche.com>
#   AuthorDate: Fri Oct 4 02:20:19 2013 -0700
#   Commit:     Adam <adam.janicki@roche.com>
#   CommitDate: Fri Oct 4 02:20:19 2013 -0700
#
#       Initial commit

# raw dump of the above history in swh format
_revisions_list = [
    {
        "author": {
            "email": b"adam.janicki@roche.com",  # noqa
            "fullname": b"Adam <adam.janicki@roche.com>",
            "name": b"Adam",
        },
        "committer": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "name": b"Adam",
        },
        "committer_date": {
            "offset_bytes": b"+0200",
            "timestamp": {"microseconds": 0, "seconds": 1380883849},
        },
        "date": {
            "offset_bytes": b"+0200",
            "timestamp": {"microseconds": 0, "seconds": 1380883849},
        },
        "directory": b"\xefX\xe7\xa6\\\xda\xdf\xfdH\xdbH\xfbq\x96@{\x98?9\xfe",
        "id": b"\xb3d\xf51U\x04NS\x08\xa0\xf7:\xbb;_\x01\x99Z[}",
        "message": b"Merge branch 'release/1.0'",
        "metadata": None,
        "parents": [
            b"\x83mI\x83\x96\xfb\x9b]E\xc8\x96\x88_\x84\xd8\xd6\nVQ\xdc",
            b"\xb9H\x86\xc5\x00\xc4n2\xdc=~\xba\xe8\xa5@\x9a\xcc\xd5\x92\xe5",
        ],  # noqa
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "name": b"Adam",
        },
        "committer": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "name": b"Adam",
        },
        "committer_date": {
            "offset_bytes": b"+0200",
            "timestamp": {"microseconds": 0, "seconds": 1380883838},
        },
        "date": {
            "offset_bytes": b"+0200",
            "timestamp": {"microseconds": 0, "seconds": 1380883838},
        },
        "directory": b"\xefX\xe7\xa6\\\xda\xdf\xfdH\xdbH\xfbq\x96@{\x98?9\xfe",
        "id": b"\xb9H\x86\xc5\x00\xc4n2\xdc=~\xba\xe8\xa5@\x9a\xcc\xd5\x92\xe5",
        "message": b"updating poms for 1.0 release",
        "metadata": None,
        "parents": [
            b"\x0c\xb6\xb4a\x1de\xbe\xe0\xf5x!\xda\xc7\xf6\x11\xe2\xf8\xa0$3"
        ],  # noqa
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "name": b"Adam",
        },
        "committer": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "name": b"Adam",
        },
        "committer_date": {
            "offset_bytes": b"+0200",
            "timestamp": {"microseconds": 0, "seconds": 1380883838},
        },
        "date": {
            "offset_bytes": b"+0200",
            "timestamp": {"microseconds": 0, "seconds": 1380883838},
        },
        "directory": b"\xefX\xe7\xa6\\\xda\xdf\xfdH\xdbH\xfbq\x96@{\x98?9\xfe",
        "id": b"\x0c\xb6\xb4a\x1de\xbe\xe0\xf5x!\xda\xc7\xf6\x11\xe2\xf8\xa0$3",
        "message": b"updating poms for 1.0 release",
        "metadata": None,
        "parents": [b"+\x02@\xc6\xd6\x82\xba\xd5\x152\xee\xc1[\x8a~\xd6\xb7\\\x8d1"],
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"janickia",
            "fullname": b"Adam Janicki <janickia>",
            "name": b"Adam Janicki",
        },
        "committer": {
            "email": b"janickia",
            "fullname": b"Adam Janicki <janickia>",
            "name": b"Adam Janicki",
        },
        "committer_date": {
            "offset_bytes": b"+0200",
            "timestamp": {"microseconds": 0, "seconds": 1380883822},
        },
        "date": {
            "offset_bytes": b"+0200",
            "timestamp": {"microseconds": 0, "seconds": 1380883822},
        },
        "directory": b"\xefX\xe7\xa6\\\xda\xdf\xfdH\xdbH\xfbq\x96@{\x98?9\xfe",
        "id": b"+\x02@\xc6\xd6\x82\xba\xd5\x152\xee\xc1[\x8a~\xd6\xb7\\\x8d1",
        "message": b"For 1.0 release. Allow untracked.\n",
        "metadata": None,
        "parents": [b"\xb4\x01\xc5\x08cG]\xb4D\x0c\x85\xc1\n\xc0\xb6B;aUM"],
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "name": b"Adam",
        },
        "committer": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "name": b"Adam",
        },
        "committer_date": {
            "offset_bytes": b"+0200",
            "timestamp": {"microseconds": 0, "seconds": 1380883692},
        },
        "date": {
            "offset_bytes": b"+0200",
            "timestamp": {"microseconds": 0, "seconds": 1380883692},
        },
        "directory": b"d@\xe7\x143w\xcb\xf7\xad\xae\x91\xd5\xec\xd8\x95\x82"
        b"\x02\xa6=\x1b",
        "id": b"\xb4\x01\xc5\x08cG]\xb4D\x0c\x85\xc1\n\xc0\xb6B;aUM",
        "message": b"updating poms for 1.0 release",
        "metadata": None,
        "parents": [b"\x9cPQ9~\\.\x0c%\x8b\xb69\xc3\xdd4@e\x84\xca\x10"],
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"janickia",
            "fullname": b"Adam Janicki <janickia>",
            "name": b"Adam Janicki",
        },
        "committer": {
            "email": b"janickia",
            "fullname": b"Adam Janicki <janickia>",
            "name": b"Adam Janicki",
        },
        "committer_date": {
            "offset_bytes": b"+0200",
            "timestamp": {"microseconds": 0, "seconds": 1380883668},
        },
        "date": {
            "offset_bytes": b"+0200",
            "timestamp": {"microseconds": 0, "seconds": 1380883668},
        },
        "directory": b"\n\x857\x94r\xbe\xcc\x04=\xe9}\xe5\xfd\xdf?nR\xe6\xa7\x9e",
        "id": b"\x9cPQ9~\\.\x0c%\x8b\xb69\xc3\xdd4@e\x84\xca\x10",
        "message": b"For 1.0 release.\n",
        "metadata": None,
        "parents": [b"\x83mI\x83\x96\xfb\x9b]E\xc8\x96\x88_\x84\xd8\xd6\nVQ\xdc"],
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"janickia",
            "fullname": b"Adam Janicki <janickia>",
            "name": b"Adam Janicki",
        },
        "committer": {
            "email": b"janickia",
            "fullname": b"Adam Janicki <janickia>",
            "name": b"Adam Janicki",
        },
        "committer_date": {
            "offset_bytes": b"+0200",
            "timestamp": {"microseconds": 0, "seconds": 1380881296},
        },
        "date": {
            "offset_bytes": b"+0200",
            "timestamp": {"microseconds": 0, "seconds": 1380881296},
        },
        "directory": b".\xf9\xa5\xcb\xb0\xd3\xdc\x9b{\xb8\x81\x03l\xe2P\x16c\x0b|\xe6",  # noqa
        "id": b"\x83mI\x83\x96\xfb\x9b]E\xc8\x96\x88_\x84\xd8\xd6\nVQ\xdc",
        "message": b"Add ignores\n",
        "metadata": None,
        "parents": [b"\xee\x96\xc2\xa2\xd3\x97\xb7\x90p\xd2\xb6\xfe0Q)\tct\x83X"],
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "name": b"Adam",
        },
        "committer": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "name": b"Adam",
        },
        "committer_date": {
            "offset_bytes": b"+0100",
            "timestamp": {"microseconds": 0, "seconds": 1380880096},
        },
        "date": {
            "offset_bytes": b"+0100",
            "timestamp": {"microseconds": 0, "seconds": 1380880096},
        },
        "directory": b"\xc7r\xc4\x9f\xc0$\xd4\xab\xff\xcb]\xf6<\xcb\x8b~\xec\xc4\xd1)",  # noqa
        "id": b"\xee\x96\xc2\xa2\xd3\x97\xb7\x90p\xd2\xb6\xfe0Q)\tct\x83X",
        "message": b"Reset author\n",
        "metadata": None,
        "parents": [b"\x8f\x89\xdd\xa8\xe0r8<\xf5\rBS*\xe8\xf5*\xd8\x9f\x8f\xdf"],
        "synthetic": False,
        "type": "git",
    },
    {
        "author": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "name": b"Adam",
        },
        "committer": {
            "email": b"adam.janicki@roche.com",
            "fullname": b"Adam <adam.janicki@roche.com>",
            "name": b"Adam",
        },
        "committer_date": {
            "offset_bytes": b"-0700",
            "timestamp": {"microseconds": 0, "seconds": 1380878419},
        },
        "date": {
            "offset_bytes": b"-0700",
            "timestamp": {"microseconds": 0, "seconds": 1380878419},
        },
        "directory": b"WS\xbaX\xd6x{q\x8f\x020i\xc5\x95\xa01\xf7y\xb2\x80",
        "id": b"\x8f\x89\xdd\xa8\xe0r8<\xf5\rBS*\xe8\xf5*\xd8\x9f\x8f\xdf",
        "message": b"Initial commit\n",
        "metadata": None,
        "parents": [],
        "synthetic": False,
        "type": "git",
    },
]


_rev_start = "b364f53155044e5308a0f73abb3b5f01995a5b7d"

_rev_missing = "836d498396fb9b5d45c896885f84d8d60a5651dc"


def check_revisions_ordering(
    swh_storage,
    rev_walker_type,
    expected_result,
    truncated_history,
    revisions_list=_revisions_list,
):
    swh_storage.revision_add(list(map(Revision.from_dict, revisions_list)))
    revision_get = swh_storage.revision_get
    swh_storage.revision_get = lambda revision_ids, ignore_displayname: [
        (
            None
            if truncated_history and rev and hash_to_hex(rev.id) == _rev_missing
            else rev
        )
        for rev in revision_get(revision_ids, ignore_displayname)
    ]

    revs_walker = get_revisions_walker(
        rev_walker_type, swh_storage, hash_to_bytes(_rev_start)
    )

    assert list(map(hash_to_bytes, expected_result)) == [
        rev["id"] for rev in revs_walker
    ]

    assert revs_walker.is_history_truncated() == truncated_history

    if truncated_history:
        missing_revs = revs_walker.missing_revisions()
        assert missing_revs == {hash_to_bytes(_rev_missing)}
    else:
        assert revs_walker.missing_revisions() == set()


def test_revisions_walker_committer_date(swh_storage):
    # revisions should be returned in reverse chronological order
    # of their committer date
    expected_result = [
        "b364f53155044e5308a0f73abb3b5f01995a5b7d",
        "b94886c500c46e32dc3d7ebae8a5409accd592e5",
        "0cb6b4611d65bee0f57821dac7f611e2f8a02433",
        "2b0240c6d682bad51532eec15b8a7ed6b75c8d31",
        "b401c50863475db4440c85c10ac0b6423b61554d",
        "9c5051397e5c2e0c258bb639c3dd34406584ca10",
        "836d498396fb9b5d45c896885f84d8d60a5651dc",
        "ee96c2a2d397b79070d2b6fe3051290963748358",
        "8f89dda8e072383cf50d42532ae8f52ad89f8fdf",
    ]

    check_revisions_ordering(
        swh_storage, "committer_date", expected_result, truncated_history=False
    )


def test_revisions_walker_dfs(swh_storage):
    # revisions should be returned in the same order they are
    # visited when performing a depth-first search in pre order
    # on the revisions DAG
    expected_result = [
        "b364f53155044e5308a0f73abb3b5f01995a5b7d",
        "836d498396fb9b5d45c896885f84d8d60a5651dc",
        "ee96c2a2d397b79070d2b6fe3051290963748358",
        "8f89dda8e072383cf50d42532ae8f52ad89f8fdf",
        "b94886c500c46e32dc3d7ebae8a5409accd592e5",
        "0cb6b4611d65bee0f57821dac7f611e2f8a02433",
        "2b0240c6d682bad51532eec15b8a7ed6b75c8d31",
        "b401c50863475db4440c85c10ac0b6423b61554d",
        "9c5051397e5c2e0c258bb639c3dd34406584ca10",
    ]

    check_revisions_ordering(
        swh_storage, "dfs", expected_result, truncated_history=False
    )


def test_revisions_walker_dfs_post(swh_storage):
    # revisions should be returned in the same order they are
    # visited when performing a depth-first search in post order
    # on the revisions DAG
    expected_result = [
        "b364f53155044e5308a0f73abb3b5f01995a5b7d",
        "b94886c500c46e32dc3d7ebae8a5409accd592e5",
        "0cb6b4611d65bee0f57821dac7f611e2f8a02433",
        "2b0240c6d682bad51532eec15b8a7ed6b75c8d31",
        "b401c50863475db4440c85c10ac0b6423b61554d",
        "9c5051397e5c2e0c258bb639c3dd34406584ca10",
        "836d498396fb9b5d45c896885f84d8d60a5651dc",
        "ee96c2a2d397b79070d2b6fe3051290963748358",
        "8f89dda8e072383cf50d42532ae8f52ad89f8fdf",
    ]

    check_revisions_ordering(
        swh_storage, "dfs_post", expected_result, truncated_history=False
    )


def test_revisions_walker_bfs(swh_storage):
    # revisions should be returned in the same order they are
    # visited when performing a breadth-first search on the
    # revisions DAG
    expected_result = [
        "b364f53155044e5308a0f73abb3b5f01995a5b7d",
        "836d498396fb9b5d45c896885f84d8d60a5651dc",
        "b94886c500c46e32dc3d7ebae8a5409accd592e5",
        "ee96c2a2d397b79070d2b6fe3051290963748358",
        "0cb6b4611d65bee0f57821dac7f611e2f8a02433",
        "8f89dda8e072383cf50d42532ae8f52ad89f8fdf",
        "2b0240c6d682bad51532eec15b8a7ed6b75c8d31",
        "b401c50863475db4440c85c10ac0b6423b61554d",
        "9c5051397e5c2e0c258bb639c3dd34406584ca10",
    ]

    check_revisions_ordering(
        swh_storage, "bfs", expected_result, truncated_history=False
    )


def test_revisions_walker_truncated_history(swh_storage):
    expected_result = [
        "b364f53155044e5308a0f73abb3b5f01995a5b7d",
        "b94886c500c46e32dc3d7ebae8a5409accd592e5",
        "0cb6b4611d65bee0f57821dac7f611e2f8a02433",
        "2b0240c6d682bad51532eec15b8a7ed6b75c8d31",
        "b401c50863475db4440c85c10ac0b6423b61554d",
        "9c5051397e5c2e0c258bb639c3dd34406584ca10",
    ]

    for revs_walker_type in ("committer_date", "bfs", "dfs", "dfs_post"):
        check_revisions_ordering(
            swh_storage, revs_walker_type, expected_result, truncated_history=True
        )


def test_revisions_walker_no_committer_date(swh_storage):
    expected_result = [
        "b364f53155044e5308a0f73abb3b5f01995a5b7d",
        "b94886c500c46e32dc3d7ebae8a5409accd592e5",
        "0cb6b4611d65bee0f57821dac7f611e2f8a02433",
        "2b0240c6d682bad51532eec15b8a7ed6b75c8d31",
        "b401c50863475db4440c85c10ac0b6423b61554d",
        "9c5051397e5c2e0c258bb639c3dd34406584ca10",
        "836d498396fb9b5d45c896885f84d8d60a5651dc",
        "ee96c2a2d397b79070d2b6fe3051290963748358",
        "8f89dda8e072383cf50d42532ae8f52ad89f8fdf",
    ]

    revisions_list = deepcopy(_revisions_list)
    for revision in revisions_list:
        revision["committer_date"] = None

    check_revisions_ordering(
        swh_storage,
        "committer_date",
        expected_result,
        truncated_history=False,
        revisions_list=revisions_list,
    )



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import attr

from swh.core.api.classes import stream_results
from swh.model.model import Person
from swh.storage.proxies.masking.db import MaskingAdmin
from swh.storage.tests.storage_data import StorageData


def test_revision_get(swh_storage, masking_admin: MaskingAdmin):
    revision = StorageData.revision
    revision2 = StorageData.revision2
    revision3 = StorageData.revision4  # StorageData.rev3 has same author/committer as 2

    assert (
        revision.author is not None
        and revision.author.email is not None
        and revision2.committer is not None
        and revision2.committer.email is not None
    )  # for mypy

    swh_storage.revision_add([revision, revision2, revision3])
    assert swh_storage.revision_get([revision.id, revision2.id, revision3.id]) == [
        revision,
        revision2,
        revision3,
    ]

    display_name1 = b"New Author <new-author1@example.org>"
    display_name2 = b"New Author <new-author2@example.org>"
    masking_admin.set_display_name(revision.author.email, display_name1)
    masking_admin.set_display_name(revision2.committer.email, display_name2)

    assert swh_storage.revision_get([revision.id, revision2.id, revision3.id]) == [
        attr.evolve(revision, author=Person.from_fullname(display_name1)),
        attr.evolve(revision2, committer=Person.from_fullname(display_name2)),
        revision3,
    ]


def test_revision_get_none_author(swh_storage, masking_admin: MaskingAdmin):
    revision = attr.evolve(
        StorageData.revision,
        author=None,
        date=None,
        committer=None,
        committer_date=None,
    )
    revision2 = StorageData.revision2
    assert revision2.author is not None and revision2.committer is not None  # for mypy
    revision2 = attr.evolve(
        revision2,
        author=attr.evolve(revision2.author, email=None),
        committer=attr.evolve(revision2.committer, email=None),
    )

    swh_storage.revision_add([revision, revision2])
    assert swh_storage.revision_get([revision.id, revision2.id]) == [
        revision,
        revision2,
    ]

    display_name = b"New Author <new-author@example.org>"
    masking_admin.set_display_name(b"old-author@example.org", display_name)

    assert swh_storage.revision_get([revision.id, revision2.id]) == [
        revision,
        revision2,
    ]


def test_revision_log(swh_storage, masking_admin: MaskingAdmin):
    revision = StorageData.revision
    revision2 = StorageData.revision2
    revision3 = StorageData.revision3
    revision4 = StorageData.revision4

    assert (
        revision.author is not None
        and revision.author.email is not None
        and revision2.committer is not None
        and revision2.committer.email is not None
    )  # for mypy

    swh_storage.revision_add([revision, revision2, revision3, revision4])
    assert list(swh_storage.revision_log([revision3.id, revision4.id])) == [
        revision3.to_dict(),
        revision.to_dict(),
        revision2.to_dict(),
        revision4.to_dict(),
    ]

    display_name1 = b"New Author <new-author1@example.org>"
    display_name2 = b"New Author <new-author2@example.org>"
    masking_admin.set_display_name(revision.author.email, display_name1)
    masking_admin.set_display_name(revision2.committer.email, display_name2)

    assert list(swh_storage.revision_log([revision3.id, revision4.id])) == [
        attr.evolve(revision3, committer=Person.from_fullname(display_name2)).to_dict(),
        attr.evolve(revision, author=Person.from_fullname(display_name1)).to_dict(),
        attr.evolve(revision2, committer=Person.from_fullname(display_name2)).to_dict(),
        revision4.to_dict(),
    ]


def test_revision_log_none_author(swh_storage, masking_admin: MaskingAdmin):
    revision = attr.evolve(
        StorageData.revision,
        author=None,
        date=None,
        committer=None,
        committer_date=None,
    )
    revision2 = StorageData.revision2
    revision3 = StorageData.revision3
    assert revision2.author is not None and revision2.committer is not None  # for mypy
    revision2 = attr.evolve(
        revision2,
        author=attr.evolve(revision2.author, email=None),
        committer=attr.evolve(revision2.committer, email=None),
    )

    swh_storage.revision_add([revision, revision2, revision3])
    assert list(swh_storage.revision_log([revision3.id])) == [
        revision3.to_dict(),
        revision.to_dict(),
        revision2.to_dict(),
    ]

    display_name = b"New Author <new-author@example.org>"
    masking_admin.set_display_name(b"old-author@example.org", display_name)

    assert list(swh_storage.revision_log([revision3.id])) == [
        revision3.to_dict(),
        revision.to_dict(),
        revision2.to_dict(),
    ]


def test_revision_get_partition(swh_storage, masking_admin: MaskingAdmin):
    revision = attr.evolve(StorageData.revision, metadata=None)
    revision2 = StorageData.revision2
    revision3 = StorageData.revision4  # StorageData.rev3 has same author/committer as 2

    assert (
        revision.author is not None
        and revision.author.email is not None
        and revision2.committer is not None
        and revision2.committer.email is not None
    )  # for mypy

    swh_storage.revision_add([revision, revision2, revision3])
    assert set(
        stream_results(
            swh_storage.revision_get_partition, partition_id=0, nb_partitions=1
        )
    ) == {
        revision,
        revision2,
        revision3,
    }

    display_name1 = b"New Author <new-author1@example.org>"
    display_name2 = b"New Author <new-author2@example.org>"
    masking_admin.set_display_name(revision.author.email, display_name1)
    masking_admin.set_display_name(revision2.committer.email, display_name2)

    assert set(
        stream_results(
            swh_storage.revision_get_partition, partition_id=0, nb_partitions=1
        )
    ) == {
        attr.evolve(revision, author=Person.from_fullname(display_name1)),
        attr.evolve(revision2, committer=Person.from_fullname(display_name2)),
        revision3,
    }


def test_revision_get_partition_none_author(swh_storage, masking_admin: MaskingAdmin):
    revision = attr.evolve(
        StorageData.revision,
        author=None,
        date=None,
        committer=None,
        committer_date=None,
        metadata=None,
    )
    revision2 = StorageData.revision2
    assert revision2.author is not None and revision2.committer is not None  # for mypy
    revision2 = attr.evolve(
        revision2,
        author=attr.evolve(revision2.author, email=None),
        committer=attr.evolve(revision2.committer, email=None),
    )

    swh_storage.revision_add([revision, revision2])
    assert set(
        stream_results(
            swh_storage.revision_get_partition, partition_id=0, nb_partitions=1
        )
    ) == {
        revision,
        revision2,
    }

    display_name = b"New Author <new-author@example.org>"
    masking_admin.set_display_name(b"old-author@example.org", display_name)

    assert set(
        stream_results(
            swh_storage.revision_get_partition, partition_id=0, nb_partitions=1
        )
    ) == {
        revision,
        revision2,
    }


############


def test_release_get(swh_storage, masking_admin: MaskingAdmin):
    release = StorageData.release
    release2 = StorageData.release2

    assert release.author is not None and release.author.email is not None  # for mypy

    swh_storage.release_add([release, release2])
    assert swh_storage.release_get([release.id, release2.id]) == [
        release,
        release2,
    ]

    display_name1 = b"New Author <new-author1@example.org>"
    masking_admin.set_display_name(release.author.email, display_name1)

    assert swh_storage.release_get([release.id, release2.id]) == [
        attr.evolve(release, author=Person.from_fullname(display_name1)),
        release2,
    ]


def test_release_get_none_author(swh_storage, masking_admin: MaskingAdmin):
    release = attr.evolve(
        StorageData.release,
        author=None,
        date=None,
    )
    release2 = StorageData.release2
    assert release2.author is not None  # for mypy
    release2 = attr.evolve(
        release2,
        author=attr.evolve(release2.author, email=None),
    )

    swh_storage.release_add([release, release2])
    assert swh_storage.release_get([release.id, release2.id]) == [
        release,
        release2,
    ]

    display_name = b"New Author <new-author@example.org>"
    masking_admin.set_display_name(b"old-author@example.org", display_name)

    assert swh_storage.release_get([release.id, release2.id]) == [
        release,
        release2,
    ]


def test_release_get_partition(swh_storage, masking_admin: MaskingAdmin):
    release = attr.evolve(StorageData.release, metadata=None)
    release2 = StorageData.release2

    assert release.author is not None and release.author.email is not None  # for mypy

    swh_storage.release_add([release, release2])
    assert set(
        stream_results(
            swh_storage.release_get_partition, partition_id=0, nb_partitions=1
        )
    ) == {
        release,
        release2,
    }

    display_name1 = b"New Author <new-author1@example.org>"
    masking_admin.set_display_name(release.author.email, display_name1)

    assert set(
        stream_results(
            swh_storage.release_get_partition, partition_id=0, nb_partitions=1
        )
    ) == {
        attr.evolve(release, author=Person.from_fullname(display_name1)),
        release2,
    }


def test_release_get_partition_none_author(swh_storage, masking_admin: MaskingAdmin):
    release = attr.evolve(
        StorageData.release,
        author=None,
        date=None,
        metadata=None,
    )

    swh_storage.release_add([release])
    assert set(
        stream_results(
            swh_storage.release_get_partition, partition_id=0, nb_partitions=1
        )
    ) == {
        release,
    }

    display_name = b"New Author <new-author@example.org>"
    masking_admin.set_display_name(b"old-author@example.org", display_name)

    assert set(
        stream_results(
            swh_storage.release_get_partition, partition_id=0, nb_partitions=1
        )
    ) == {
        release,
    }


def test_set_display_names(swh_storage, masking_admin: MaskingAdmin):
    def get_all():
        cur = masking_admin.cursor()
        cur.execute("SELECT * FROM display_name")
        return set(cur.fetchall())

    display_names = {
        (
            f"Author {i} <author{i}@example.org>".encode(),
            f"New Author {i} <new-author{i}example.org>".encode(),
        )
        for i in range(10)
    }

    masking_admin.set_display_names(display_names)
    assert get_all() == display_names

    # doing it again with clear=True should not change anything
    masking_admin.set_display_names(display_names, clear=True)
    assert get_all() == display_names

    # doing it again should not change anything
    masking_admin.set_display_names(display_names)
    assert get_all() == display_names

    more_display_names = {
        (
            f"Person {i} <person{i}@example.org>".encode(),
            f"New Person {i} <new-person{i}example.org>".encode(),
        )
        for i in range(10)
    }

    masking_admin.set_display_names(more_display_names)
    assert get_all() == (display_names | more_display_names)

    masking_admin.set_display_names(more_display_names)
    assert get_all() == (display_names | more_display_names)

    # with clear=True, should only get the inserted objects
    masking_admin.set_display_names(display_names, clear=True)
    assert get_all() == display_names



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import dataclasses
import enum
import graphlib
import logging
import textwrap
from typing import Callable, Iterable, Optional, Sequence

from .cql import CqlRunner, create_table
from .model import MigrationRow

logger = logging.getLogger(__name__)


class MigrationStatus(enum.Enum):
    PENDING = "pending"
    """The migration was not applied yet"""
    RUNNING = "running"
    COMPLETED = "completed"


@dataclasses.dataclass
class Migration:
    id: str
    """Unique identifier of this migration.

    Should have the format: ``YYYY-MM-DD_developer_readable_name``"""

    dependencies: set[str]
    """Set of identifiers of migrations this migration depends on"""

    min_read_version: str
    """Lowest version of the Python code that should be allowed to read the database
    if this migration is applied"""

    script: Optional[Callable[[CqlRunner], None]]
    """If provided, this is a function that runs the migration.

    If not provided, the migration must be run manually, using steps described
    in the documentation"""

    help: Optional[str]
    """Documentation of the migration

    Typically describes what to do if ``script`` is :const:`None`."""

    required: bool
    """Whether this migration must be applied for the current version of the Python code
    to allow instantiating :class:`swh.storage.cassandra.CassandraStorage`."""


MIGRATIONS: tuple[Migration, ...] = (
    Migration(
        id="2024-12-12_init",
        dependencies=set(),
        min_read_version="2.9.0",
        script=lambda _cql_runner: None,
        help="Dummy migration that represents the database schema as of v2.9.0" "",
        required=True,
    ),
)


def list_migrations(
    cql_runner: CqlRunner, rows: Optional[Sequence[MigrationRow]] = None
) -> list[tuple[Migration, MigrationStatus]]:
    """Returns all known migrations, in topological order

    ``rows``, should be the value returned by ``cql_runner.migration_list``.

    This includes migrations that are not required to instantiate
    :class:`swh.storage.cassandra.CassandraStorage`."""
    dependency_graph = {m.id: m.dependencies for m in MIGRATIONS}
    if rows is None:
        rows = list(cql_runner.migration_list())
    statuses = {row.id: row.status for row in rows}
    migrations = {migration.id: migration for migration in MIGRATIONS}
    return [
        (
            migrations[migration_id],
            MigrationStatus(statuses.get(migration_id, "pending")),
        )
        for migration_id in graphlib.TopologicalSorter(dependency_graph).static_order()
    ]


def apply_migrations(
    cql_runner: CqlRunner, ids_to_apply: Iterable[str]
) -> tuple[bool, Sequence[Migration], Sequence[Migration]]:
    """Applies migrations with the given ids (unless they already are).

    Returns:
        * whether any was run, and
        * which migrations still need to be run manually.
        * which migrations cannot run because they are missing dependencies
    """
    applied_any = False
    remaining_manual_migrations = []
    remaining_migrations_missing_dependencies = []

    statuses = {
        migration.id: status for (migration, status) in list_migrations(cql_runner)
    }
    for migration_id in ids_to_apply:
        if migration_id not in statuses:
            raise ValueError(f"Unknown migration: {migration_id}")

    migrations_to_apply = [
        migration for migration in MIGRATIONS if migration.id in ids_to_apply
    ]
    for migration in migrations_to_apply:
        status = statuses[migration.id]
        if status == MigrationStatus.PENDING:
            missing_dependencies = {
                dependency
                for dependency in migration.dependencies
                if statuses[dependency] != MigrationStatus.COMPLETED
            }
            if missing_dependencies:
                logger.warning(
                    "Cannot apply %s: depends on %s",
                    migration.id,
                    ", ".join(missing_dependencies),
                )
                remaining_migrations_missing_dependencies.append(migration)
                continue
            cql_runner.migration_add_one(
                MigrationRow(
                    id=migration.id,
                    dependencies=migration.dependencies,
                    min_read_version=migration.min_read_version,
                    status=MigrationStatus.RUNNING.value,
                )
            )
            if migration.script is None:
                logger.info("Skipping %s", migration.id)
                if migration.help:
                    logger.info("%s", textwrap.indent(migration.help, "    "))
                remaining_manual_migrations.append(migration)
            else:
                logger.info("Running %s...", migration.id)
                migration.script(cql_runner)
                cql_runner.migration_add_one(
                    MigrationRow(
                        id=migration.id,
                        dependencies=migration.dependencies,
                        min_read_version=migration.min_read_version,
                        status=MigrationStatus.COMPLETED.value,
                    )
                )
                logger.info("Done.")
                statuses[migration.id] = MigrationStatus.COMPLETED
                applied_any = True

    return (
        applied_any,
        remaining_manual_migrations,
        remaining_migrations_missing_dependencies,
    )


def create_migrations_table_if_needed(cql_runner: CqlRunner) -> None:
    if not list(
        cql_runner.execute_with_retries(
            """
            SELECT table_name
            FROM system_schema.tables
            WHERE keyspace_name=%s AND table_name='migration'
            """,
            [cql_runner.keyspace],
        )
    ):
        logger.info("'migrations' table does not exist yet, creating it.")

        # 'migration' table does not exist. Create it:
        cql_runner.execute_with_retries(f'USE "{cql_runner.keyspace}"', [])
        create_table(cql_runner, "migration")

        # And mark the dummy initial migration as done, as it corresponds to the schema
        # of the last swh-storage version before adding the 'migrations' table.
        # Other migrations could not have run before this is done.
        (migration,) = [
            migration for migration in MIGRATIONS if migration.id == "2024-12-12_init"
        ]

        migration_row = MigrationRow(
            id=migration.id,
            dependencies=migration.dependencies,
            min_read_version=migration.min_read_version,
            status=MigrationStatus.COMPLETED.value,
        )
        cql_runner.migration_add_concurrent([migration_row])
        logger.info("'migrations' table created.")



# Copyright (C) 2020-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import logging

import pytest

from .test_cli import invoke

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    ("start", "end", "expected_weeks", "unexpected_weeks"),
    (
        pytest.param(
            "2020-02-03",
            "2020-02-09",
            [("2020w06", "2020-02-03", "2020-02-10")],
            ["2020w05", "2020w07"],
            id="single-week",
        ),
        pytest.param(
            "2023-01-01",
            "2023-01-01",
            [("2022w52", "2022-12-26", "2023-01-02")],
            ["2022w51", "2023w01"],
            id="week-in-another-year",
        ),
        pytest.param(
            "2023-01-01",
            "2023-01-17",
            [
                ("2022w52", "2022-12-26", "2023-01-02"),
                ("2023w01", "2023-01-02", "2023-01-09"),
                ("2023w02", "2023-01-09", "2023-01-16"),
                ("2023w03", "2023-01-16", "2023-01-23"),
            ],
            ["2022w51", "2023w04"],
            id="multiple-weeks",
        ),
    ),
)
def test_create_object_reference_partitions(
    swh_storage_backend_config,
    swh_storage,
    start,
    end,
    expected_weeks,
    unexpected_weeks,
):
    storage_config = {
        "storage": {
            "cls": "pipeline",
            "steps": [
                {"cls": "record_references"},
                swh_storage_backend_config,
            ],
        }
    }

    result = invoke(
        "create-object-reference-partitions",
        start,
        end,
        local_config=storage_config,
    )

    assert result.exit_code == 0, result.output

    partitions = swh_storage.object_references_list_partitions()

    for partition, (week, expected_start, expected_end) in zip(
        partitions, expected_weeks
    ):
        assert partition.table_name == f"object_references_{week}"
        assert partition.start == datetime.date.fromisoformat(expected_start)
        assert partition.end == datetime.date.fromisoformat(expected_end)

    table_names = {partition.table_name for partition in partitions}
    for week in unexpected_weeks:
        assert f"object_references_{week}" not in table_names


@pytest.fixture
def swh_storage_with_partitions(swh_storage):
    # Setup partitions from 2022-12-26 00:00 to 2023-01-15 23:59
    for partition in swh_storage.object_references_list_partitions():
        swh_storage.object_references_drop_partition(partition)
    swh_storage.object_references_create_partition(2022, 52)
    for week in range(1, 3):
        swh_storage.object_references_create_partition(2023, week)
    yield swh_storage
    for partition in swh_storage.object_references_list_partitions():
        swh_storage.object_references_drop_partition(partition)


@pytest.mark.parametrize(
    ("before", "listed_in_output", "remaining"),
    (
        pytest.param(
            "2022-12-25",
            [],
            ["2022w52", "2023w01", "2023w02"],
            id="older-than-first-week",
        ),
        pytest.param(
            "2023-01-01",
            [],
            ["2022w52", "2023w01", "2023w02"],
            id="sunday-of-first-week",
        ),
        pytest.param(
            "2023-01-02",
            ["2022 week 52"],
            ["2023w01", "2023w02"],
            id="middle-of-second-week",
        ),
        pytest.param(
            "2023-01-12",
            ["2022 week 52", "2023 week 01"],
            ["2023w02"],
            id="middle-of-third-week",
        ),
    ),
)
def test_remove_old_object_reference_partitions(
    swh_storage_backend_config,
    swh_storage_with_partitions,
    before,
    listed_in_output,
    remaining,
):
    storage_config = {
        "storage": {
            "cls": "pipeline",
            "steps": [
                {"cls": "record_references"},
                swh_storage_backend_config,
            ],
        }
    }

    result = invoke(
        "remove-old-object-reference-partitions",
        before,
        input="y\n",
        local_config=storage_config,
    )

    assert result.exit_code == 0, result.output

    assert all(s in result.output for s in listed_in_output)

    partitions = swh_storage_with_partitions.object_references_list_partitions()

    table_names = {partition.table_name for partition in partitions}

    # We get a partition created for the current week when the schema is initialized
    table_names.discard(f"object_references_{datetime.date.today().strftime('%Yw%W')}")
    assert {f"object_references_{week}" for week in remaining} == table_names


def test_remove_old_object_reference_partitions_postgresql_refuses_to_remove_all(
    swh_storage_backend_config, swh_storage_with_partitions
):
    storage_config = {
        "storage": {
            "cls": "pipeline",
            "steps": [
                {"cls": "record_references"},
                swh_storage_backend_config,
            ],
        }
    }

    result = invoke(
        "remove-old-object-reference-partitions",
        # A partition for this week is created with the schema,
        # so we just ask to delete
        (datetime.date.today() + datetime.timedelta(weeks=1)).strftime("%Y-%m-%d"),
        local_config=storage_config,
    )

    assert result.exit_code != 0
    assert "Trying to remove all existing partitions" in result.output

    partitions = swh_storage_with_partitions.object_references_list_partitions()

    # The three partition created in the fixture
    assert len(partitions) == 3


def test_remove_old_object_reference_partitions_using_force_argument(
    swh_storage_backend_config, swh_storage_with_partitions
):
    storage_config = {
        "storage": {
            "cls": "pipeline",
            "steps": [
                {"cls": "record_references"},
                swh_storage_backend_config,
            ],
        }
    }

    swh_storage_with_partitions.object_references_create_partition(
        *datetime.date.today().isocalendar()[0:2]
    )

    result = invoke(
        "remove-old-object-reference-partitions",
        "--force",
        "2023-01-31",
        local_config=storage_config,
    )

    assert result.exit_code == 0, result.output

    partitions = swh_storage_with_partitions.object_references_list_partitions()

    table_names = {partition.table_name for partition in partitions}

    assert not any(
        f"object_references_{week}" in table_names
        for week in ("2023w01", "2023w02", "2023w03")
    )



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from os import path

import swh.storage

SQL_DIR = path.join(path.dirname(swh.storage.__file__), "sql")



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools

import pytest

from swh.storage.exc import MaskedObjectException
from swh.storage.proxies.masking.cls import MaskingProxyStorage
from swh.storage.proxies.masking.db import MaskedState
from swh.storage.tests.storage_data import StorageData
from swh.storage.tests.test_in_memory import TestInMemoryStorage as _TestStorage

MASKED_SWHIDS = {
    StorageData.content.swhid().to_extended(),
    StorageData.directory.swhid().to_extended(),
    StorageData.revision.swhid().to_extended(),
    StorageData.release.swhid().to_extended(),
    StorageData.snapshot.swhid().to_extended(),
    StorageData.origin.swhid(),
}


@pytest.fixture
def swh_storage(masking_db_postgresql, masking_admin, swh_storage_backend):
    # Create a request
    request = masking_admin.create_request(slug="foo", reason="bar")

    masking_admin.set_object_state(
        request_id=request.id,
        new_state=MaskedState.DECISION_PENDING,
        swhids=list(MASKED_SWHIDS),
    )

    storage = MaskingProxyStorage(
        db=masking_db_postgresql.info.dsn, storage=swh_storage_backend
    )
    try:
        yield storage
    finally:
        storage._masking_pool.close()


class TestStorage(_TestStorage):
    @pytest.mark.xfail(reason="typing.Protocol instance check is annoying")
    def test_types(self, *args, **kwargs):
        super().test_types(*args, **kwargs)

    @pytest.mark.xfail(
        reason="This test only adds one origin with a status, and it's masked"
    )
    def test_origin_visit_status_get_random(self, *args, **kwargs):
        super().test_origin_visit_status_get_random(*args, **kwargs)


EXPECTED_MASKED_EXCEPTIONS = {
    "test_content_add",
    "test_content_add__legacy",
    "test_content_add_duplicate",
    "test_content_add_from_lazy_content",
    "test_content_add_metadata",
    "test_content_add_twice",
    "test_content_find_ctime",
    "test_content_find_with_duplicate_blake2s256",
    "test_content_find_with_duplicate_input",
    "test_content_find_with_duplicate_sha256",
    "test_content_find_with_present_content",
    "test_content_get",
    "test_content_get_data_full_dict",
    "test_content_get_data_missing",
    "test_content_get_data_single_hash_dict",
    "test_content_get_data_two_hash_dict",
    "test_content_get_missing",
    "test_content_metadata_add",
    "test_content_metadata_add_duplicate",
    "test_content_metadata_get",
    "test_content_metadata_get_after",
    "test_content_metadata_get_authorities",
    "test_content_metadata_get_by_ids",
    "test_content_metadata_get_paginate",
    "test_content_metadata_get_paginate_same_date",
    "test_directory_add",
    "test_directory_add_with_raw_manifest",
    "test_directory_get_id_partition",
    "test_directory_ls_non_recursive",
    "test_directory_ls_recursive",
    "test_extid_add_extid_multicity",
    "test_extid_add_git",
    "test_extid_add_target_multicity",
    "test_extid_add_twice",
    "test_extid_version_behavior",
    "test_origin_add",
    "test_origin_get",
    "test_origin_get_by_sha1",
    "test_origin_metadata_add",
    "test_origin_metadata_add_duplicate",
    "test_origin_metadata_get",
    "test_origin_metadata_get_after",
    "test_origin_metadata_get_paginate",
    "test_origin_metadata_get_paginate_same_date",
    "test_origin_search_multiple_visit_types",
    "test_origin_search_no_regexp",
    "test_origin_search_regexp_fullstring",
    "test_origin_search_regexp_substring",
    "test_origin_search_single_result",
    "test_origin_search_with_visit_types",
    "test_origin_snapshot_get_all",
    "test_origin_visit_find_by_date",
    "test_origin_visit_find_by_date_and_type",
    "test_origin_visit_find_by_date_latest_visit",
    "test_origin_visit_get_all",
    "test_origin_visit_get_by",
    "test_origin_visit_get_latest",
    "test_origin_visit_get_latest__not_last",
    "test_origin_visit_get_latest__same_date",
    "test_origin_visit_get_latest_filter_type",
    "test_origin_visit_get_latest_order",
    "test_origin_visit_get_with_statuses",
    "test_origin_visit_status_get_all",
    "test_origin_visit_status_get_latest",
    "test_release_get",
    "test_release_get_order",
    "test_release_get_partition",
    "test_revision_add_fractional_timezone",
    "test_revision_get",
    "test_revision_get_no_parents",
    "test_revision_get_order",
    "test_revision_get_partition",
    "test_revision_log",
    "test_revision_shortlog",
    "test_snapshot_add_get",
    "test_snapshot_add_many",
    "test_snapshot_add_many_incremental",
    "test_snapshot_branch_get_by_name_missing_branch",
    "test_snapshot_get_id_partition",
}

for method_name in dir(TestStorage):
    if not method_name.startswith("test_"):
        continue

    method = getattr(TestStorage, method_name)

    @functools.wraps(method)
    def wrapped_test(*args, method=method, **kwargs):
        try:
            method(*args, **kwargs)
        except MaskedObjectException as exc:
            assert (
                method.__name__ in EXPECTED_MASKED_EXCEPTIONS
            ), f"{method.__name__} shouldn't raise a MaskedObjectException"

            assert exc.masked, "The exception was raised with no masked objects"

            assert (
                set(exc.masked.keys()) <= MASKED_SWHIDS
            ), "Found an unexpectedly masked object"
        else:
            assert (
                method.__name__ not in EXPECTED_MASKED_EXCEPTIONS
            ), f"{method.__name__} should raise a MaskedObjectException"

    setattr(TestStorage, method_name, wrapped_test)
